/*
 * @(#) TransactionRecoveryImpl.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 *
 * This module was originally developed by 
 *
 *  - Bull S.A. as part of the JOnAS application server code released in 
 *    July 1999 (www.bull.com)
 * 
 * --------------------------------------------------------------------------
 *  The original code and portions created by Bull SA are 
 *  Copyright (c) 1999 BULL SA  
 *  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * --------------------------------------------------------------------------
 * $Id: TransactionRecoveryImpl.java,v 1.1 2004/10/29 05:48:30 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jotm;

import java.io.IOException;
import java.io.FileInputStream;
import java.nio.ByteBuffer;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;

import javax.transaction.SystemException;
import javax.transaction.xa.XAResource;

import org.objectweb.howl.log.Configuration;
import org.objectweb.howl.log.LogConfigurationException;
import org.objectweb.howl.log.LogException;
import org.objectweb.howl.log.LogRecord;
import org.objectweb.howl.log.LogRecordType;
import org.objectweb.howl.log.ReplayListener;
import org.objectweb.howl.log.xa.XACommittingTx;
import org.objectweb.howl.log.xa.XALogRecord;
import org.objectweb.howl.log.xa.XALogger;

import javax.transaction.xa.XAException;

/**
 * 
 * @author Tony Ortiz
 */

public class TransactionRecoveryImpl implements TransactionRecovery {

    private transient static TransactionRecoveryImpl unique = null;

    // Static hashtable: Resource Manager Name ---> Connection
    private transient static Map nameResourceManager = Collections.synchronizedMap(new HashMap());

    // Static hashtable: XA Resource Name ---> XAResource
    private transient static Map nameXAResource = Collections.synchronizedMap(new HashMap());

    // Jotm Recovery
    private transient static JotmRecovery tmrecovery = null;
    private transient static boolean dorecovery = false;

    // XALogger info
    private transient static XALogger xaLog = null;
    private transient static XACommittingTx rmCommitTx = null;

    // Vector used to hold all registered Resource Managers
    // until they can be written to the Howl Logger.
    private transient static Vector rmreg = new Vector();
    private RmRegistration myrmRegistration = null;
    
    /**
     * -Djotm.base property
     */
    private static final String JOTM_BASE = "jotm.base";

    /**
     * jonas.base property
     */
    private static final String JONAS_BASE = "jonas.base";
    
    /**
     * configuration directory name
     */
    private static final String CONFIG_DIR = "conf";

    /**
     * System properties
     */
    private static Properties systEnv = System.getProperties();  

    /**
     * JONAS_BASE
     */
    private static String jonasBase = systEnv.getProperty(JONAS_BASE);
    
    /**
     * JOTM_BASE
     */
    private static String jotmBase = systEnv.getProperty(JOTM_BASE);

    /**
     * Separator of file
     */
    private static String fileSeparator = systEnv.getProperty("file.separator");

    /**
     * Default constructor.
     */

    public TransactionRecoveryImpl() throws LogException, IOException, Exception {
            
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("TransactionRecoveryImpl constructor");
        }

        unique = this;
        
        // Check the JONAS_BASE and JOTM_BASE environment properties
        // If on of them is not found, check the transactionRecovery value
        // from Current to see if recovery is enabled(true)/disabled(false).
        
        String myBase=null;
        
        if (jonasBase == null) {
            if (jotmBase == null) {
                if (!Current.getCurrent().getDefaultRecovery()) {
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("JOTM Recovery is disabled");
                    }
                    return;    // transaction recovery is disabled   
                }
            } else {
               myBase = jotmBase;
            }
        } else {
           myBase = jonasBase;   
        }
        
        myBase = myBase.trim();

        // JOTM_BASE/conf/fileName
        String fileFullPathname = myBase + fileSeparator + CONFIG_DIR + fileSeparator + "jotm.properties";
        
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("JOTM properties file= " + fileFullPathname);
        }
            
        Properties howlprop = new Properties();
        try {
            FileInputStream inStr = new FileInputStream(fileFullPathname);
            systEnv.load(inStr);
        } catch (Exception e){
            Current.getCurrent().setDefaultRecovery(false);
            return;   
        }
        
        if (systEnv.getProperty("jotm.recovery.Enabled").trim().equalsIgnoreCase("true")) {    // recovery is enabled
            Current.getCurrent().setDefaultRecovery(true);
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is enabled");
            }
        } else {    // recovery disabled
            Current.getCurrent().setDefaultRecovery(false);
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return;
        }
        
        String myhowlprop = null;
        myhowlprop = systEnv.getProperty ("howl.log.ListConfiguration", "false");
        howlprop.put("listConfig", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.BufferSize", "4");
        howlprop.put("bufferSize", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.MinimumBuffers", "16");
        howlprop.put("minBuffers", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.MaximumBuffers", "16");
        howlprop.put("maxBuffers", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.MaximumBlocksPerFile", "200");
        howlprop.put("maxBlocksPerFile", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.FileDirectory", systEnv.getProperty ("basedir", "."));
        howlprop.put("logFileDir", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.FileName", "howl");
        howlprop.put("logFileName", myhowlprop);
        myhowlprop = systEnv.getProperty ("howl.log.MaximumFiles", "2");
        howlprop.put("maxLogFiles", myhowlprop);
        
        try {
            howlOpenLog (howlprop);
        } catch (Exception e) {
            TraceTm.jotm.error("howlOpenLog: LogException");
            throw new SystemException("LogException occured in howlOpenLog() " + e.getMessage());
        }
    }

    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized in case of plain client.
     *
     * @return The <code>Current</code> object created 
     */

    public static TransactionRecoveryImpl getTransactionRecovery() {

        return unique;
    }

    // ------------------------------------------------------------------
    // JOTM Recovery Support Methods
    // ------------------------------------------------------------------

    /**
     * Register a Resource Manager with the JOTM Transaction Manager.
     * 
     * @param rmName The Resource Manager to be registered.
     */

    public void registerResourceManager (String rmName, XAResource rmXares, String info,
                                         TransactionResourceManager tranrm) throws XAException {
        
        if (!(tranrm==null)) {  // initial implementation, call back immediately
            tranrm.returnXAResource(rmName, rmXares);   
        }
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return;    // transaction recovery is disabled   
        }
        
        // put the Resource Manager/XAResource mapping into the hashtable

        if ( TraceTm.recovery.isDebugEnabled() ) {
            TraceTm.recovery.debug("Register Resource Manager " + rmName + " to Connection " + rmXares);
        }

        XAResource xares = null;
        Object hash_key = new String(rmName);

        Object myrmXares = (Object) nameResourceManager.get(hash_key);

        if (myrmXares == null) {
            nameResourceManager.put(hash_key, rmXares);
        }
        else {
        	xares = (XAResource) myrmXares;
            if (xares.equals(rmXares)) {
                if( TraceTm.recovery.isDebugEnabled() ) {
                    TraceTm.recovery.debug(rmName + " already registered");
                    return;
                }
            }
            else {
                nameResourceManager.put(hash_key, rmXares);
            }
        }

        myrmRegistration = new RmRegistration();
        myrmRegistration.rmAddRegistration (rmName, rmXares);
        rmreg.addElement(myrmRegistration);
        
    }

    /**
     * Provide information regarding the status and state of the XAResource.
     * 
     * @param rmName The Resource Manager to be reported upon.
     *
     * @return XAResource The XAResource assigned to the Resource Managere.
     */

    public XAResource reportResourceManager (String rmName) throws XAException {
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return null;    // transaction recovery is disabled   
        }
        
        // given the name, get the corresponding Connection from the hashtable.

        if( TraceTm.recovery.isDebugEnabled() ) {
            TraceTm.recovery.debug("get Connection from Resource Manager " + rmName);
        }

        Object hash_key = new String(rmName);
        XAResource myXares = (XAResource) nameResourceManager.get(hash_key);

        if (myXares == null) {
            throw new XAException ("Named Resource Manager " + rmName + " does not exist");
        }

        return myXares;
    }

    /**
     * Unregister a Resource Manager from the JOTM Transaction Manager.
     * 
     * @param rmName The Resource Manager to be unregistered.
     */

    public void unregisterResourceManager(String rmName, XAResource rmXares) throws XAException {
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return;    // transaction recovery is disabled   
        }
        
        // Unregister the Resource Manager from the name (hash table).

        if ( TraceTm.recovery.isDebugEnabled() ) {
            TraceTm.recovery.debug("Remove Resource Manager " + rmName + " from Connection " + rmXares);
        }

        Object hash_key = new String(rmName);
        XAResource myrmXares = (XAResource) nameResourceManager.get(hash_key);

        if (myrmXares.equals(rmXares)) {
            nameResourceManager.remove(hash_key);
        }
        else {
            throw new XAException ("Resource Manager " + rmName + " not associated to " + rmXares);
        }

        // try {
        //     Method m = rmObj.getClass().getMethod("freeXAResource", new Class [] {XAResource.class} );
        //     m.invoke (rmObj, new Object [] {invokexares} );
        // } catch (InvocationTargetException e) {
        //     throw new XAException("Cannot invoke method " + e.getMessage());
        // } catch (NoSuchMethodException e) {
        //     throw new XAException("Method does not exist " + e.getMessage());
        // } catch (IllegalAccessException e ) {
        //     throw new XAException("Cannot access method " + e.getMessage());
        // }

    }
    
    /**
     * Log (in Howl) every Resource Manager (XAResource) that has been
     * registered.
     * 
     * @exception XAException Thrown if the transaction manager 
     * encounters an unexpected error condition
     */

    public void startResourceManagerRecovery () throws XAException {
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return;    // transaction recovery is disabled   
        }

        int rmcount = rmreg.size();
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("LogResourceManager count= " +rmcount);
        }

        if (rmcount == 0) return;

        XACommittingTx myrmCommitTx = null;

        byte [] [] rmBuffer = new byte [rmcount + 1] [];

        byte [] rmRecord1 = null;
        byte [] rmRecord2 = null;

        String resm1 = "RM1";
        String resm2 = "RM2";

        long rmdate = System.currentTimeMillis();

        rmRecord1 = new byte[3+8+4];

        ByteBuffer rm1 = ByteBuffer.wrap(rmRecord1);
        rm1.put(resm1.getBytes());
        rm1.putLong(rmdate);
        rm1.putInt(rmcount);

        rmBuffer [0] = rm1.array();

        for (int i = 0; i < rmcount; i++) {
            myrmRegistration = (RmRegistration)rmreg.elementAt(i);
            String rmName = myrmRegistration.rmGetName();
            XAResource xaRes = myrmRegistration.rmGetXaRes();

            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("LogResourceManager rmName= " + rmName);
                TraceTm.recovery.debug("    xaRes= ");
                TraceTm.recovery.debug("    rmIndex= " + i);
            }

            int rmlength = rmName.length();
            int xaReslength= xaRes.toString().length();

            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("rm length=" + rmlength);
                TraceTm.recovery.debug("xaRes length= " + xaReslength);
            }

            rmRecord2 = new byte[3+4+rmlength+4+xaReslength+4];
            ByteBuffer rm2 = ByteBuffer.wrap(rmRecord2);

            rm2.put(resm2.getBytes());
            rm2.putInt(rmlength);
            rm2.put(rmName.getBytes());
            rm2.putInt(xaReslength);
            rm2.put(xaRes.toString().getBytes());
            rm2.putInt(i);

            rmBuffer [i+1] = rm2.array();    // First record (0) is always rm1
        }

        try {
            myrmCommitTx = howlCommitLog(rmBuffer);
        } catch (Exception e) {
            // If we cannot write the Log, we cannot perform recovery

            String howlerror =
                "Cannot howlCommitLog:"
                + e
                + " --"
                + e.getMessage();
            TraceTm.jotm.error("Got LogException from howlCommitLog: "+ howlerror);
            rmCommitTx = null;

            throw new XAException(howlerror);
        }

        if (!(rmCommitTx == null)) {
            byte [] rmDone = new byte [11];
            byte [] [] rmDoneRecord = new byte [1] [11];

            rmDone = "RM3JOTMDONE".getBytes();

            try {
                rmDoneRecord [0] = rmDone;
                howlDoneLog (rmDoneRecord, rmCommitTx);
            } catch (Exception f) {
                String howlerror =
                    "Cannot howlDoneLog:"
                    + f
                    + "--"
                    + f.getMessage();
                TraceTm.jotm.error("Got LogException from howlDoneLog: "+ howlerror);
            }
        }

        rmCommitTx = myrmCommitTx;
        
        try {
            recoverResourceManager();
        } catch (XAException e) {
            throw new XAException("Cannot perform recovery " + e.getMessage());
        }
    }
    
    /**
     * Recover a Resource Manager with the JOTM Transaction Manager.
     * 
     * @exception XAException Thrown if the transaction manager 
     * encounters an unexpected error condition
     */

    public void recoverResourceManager () throws XAException {
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }
            return;    // transaction recovery is disabled   
        }
        
        if ( TraceTm.recovery.isDebugEnabled() ) {
            TraceTm.recovery.debug("recoverResourceManager");
        }

        if (rmreg.size()== 0){
            if ( TraceTm.recovery.isDebugEnabled() ) {
                TraceTm.recovery.debug("Nothing to recover");
            }
            return;
        }
        
        // If incomplete commit records were found in Howl, we may need to
        // recover transactions (xids).

        // if (dorecovery) {
            try {
                tmrecovery.recoverTransactions(rmreg);
            } catch (XAException e) {
                throw new XAException("Unable to recover transactions" + e.getMessage());
            }
        // } else {
        //     if ( TraceTm.recovery.isDebugEnabled() ) {
        //         TraceTm.recovery.debug("Nothing to recover");
        //     }
        // }
    }

    /* HOWL Support Methods */
    
    private class xaReplayListener implements ReplayListener {

        public void onRecord (LogRecord lr) {

            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("LogRecord type= " + lr.type);
            }

            switch(lr.type) {
                case LogRecordType.EOB:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Howl End of Buffer Record");
                    }
                    break;
                case LogRecordType.END_OF_LOG:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Howl End of Log Record");
                    }
                    break;
                case LogRecordType.XACOMMIT:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Howl XA Commit Record");
                    }
                    dorecovery = tmrecovery.rebuildTransaction ((XALogRecord) lr);
                    break;
                case LogRecordType.XADONE:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Howl XA Done Record");
                    }
                    break;
                case LogRecordType.USER:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Howl User Record");
                    }
                    break;
                default:
                    if (TraceTm.recovery.isDebugEnabled()) {
                        TraceTm.recovery.debug("Unknown Howl LogRecord");
                    }
                    break;
            }
        }

        public void onError(LogException exception) {
          // TODO Auto-generated method stub
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("onError");
            }
        }

        public LogRecord getLogRecord() {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("getLogRecord");
            }
            return new XALogRecord(120);
        }
    }
   
    /**
     * open  the Howl Log
     */

    synchronized void howlOpenLog (Properties phowlprop) throws SystemException {

        if (!(xaLog == null)) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Howl Log already opened");
            }
            return;
        }

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Open howl log");
        }

        try {
            Configuration cfg = new Configuration (phowlprop);
            xaLog = new XALogger(cfg);
        } catch (LogConfigurationException e) {
            TraceTm.jotm.error("XALogger: LogConfigurationException");
            throw new SystemException("LogConfigurationException occured in XALogger() " + e.getMessage());
        } catch (IOException e) {
            TraceTm.jotm.error("XALogger: IOException");
            throw new SystemException("IOException occured in XALogger() " + e.getMessage());
        } catch (Exception e) {
            TraceTm.jotm.error("XALogger: Exception");
            throw new SystemException("Exeception occurred in XALogger() " + e.getMessage());
        }

        tmrecovery = new JotmRecovery ();

        xaReplayListener myxarl = new xaReplayListener();

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("xaLog.open");
        }

        try {
            xaLog.open (null);
            // xaLog.open (myxarl);
        } catch (LogException e) {
            TraceTm.jotm.error("xaLog.open: LogException");
            throw new SystemException("LogException occured in xaLog.open() " + e.getMessage());
        } catch (IOException e) {
            TraceTm.jotm.error("xaLog.open: IOException");
            throw new SystemException("IOException occured in xaLog.open() " + e.getMessage());
        } catch (InterruptedException e) {
            TraceTm.jotm.error("xaLog.open: InterruptedException");
            throw new SystemException("InterruptedException occured in xaLog.open() " + e.getMessage());
        } catch (ClassNotFoundException e) {
            TraceTm.jotm.error("xaLog.open: ClassNotFoundException");
            throw new SystemException("ClassNotFoundException occured in xaLog.open() " + e.getMessage());
        } catch (Exception e) {
            TraceTm.jotm.error("xaLog.open: Exception " + e.getMessage());
            throw new SystemException("Exception occurred in xaLog.open() " + e.getMessage());
        }

        xaLog.replayActiveTx (myxarl);
    }

    /**
     * write the Done record to the Howl Log
     */

    void howlCloseLog () throws SystemException {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Close howl log");
        }

        if (xaLog == null) return;

        // We maybe able to write a RM3 Done record if no errors occurred
        // within JOTM.
        
        // if (!(rmCommitTx == null)) {
        //     byte [] rmDone = new byte [11];
        //     byte [] [] rmDoneRecord = new byte [1] [11];

        //     rmDone = "RM3JOTMDONE".getBytes();
        //     try {
        //         rmDoneRecord [0] = rmDone;
        //         howlDoneLog (rmDoneRecord, rmCommitTx);
        //         rmCommitTx = null;
        //     } catch (Exception f) {
        //         String howlerror =
        //             "Cannot howlDoneLog:"
        //             + f
        //             + "--"
        //             + f.getMessage();
        //         TraceTm.jotm.error("Got LogException from howlDoneLog: "+ howlerror);
        //     }
        // }

        try {
            xaLog.close();
        } catch (LogException e) {
            TraceTm.jotm.error("xaLog.close: LogException");
            throw new SystemException("LogException occured in xaLog.close() " + e.getMessage());
        } catch (IOException e) {
            TraceTm.jotm.error("xaLog.close: IOException");
            throw new SystemException("IOException occured in xaLog.close() " + e.getMessage());
        } catch (InterruptedException e) {
            TraceTm.jotm.error("xaLog.close: InterruptedException");
            throw new SystemException("InterruptedException occured in xaLog.close() " + e.getMessage());
        }

        xaLog = null;

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Howl log closed");
        }
    }

    /**
     * write the Commit record to the Howl Log
     */

    public XACommittingTx howlCommitLog (byte [] [] xaCmRec) throws LogException, Exception {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Commit howl log");
        }
        return xaLog.putCommit (xaCmRec);
    }

    /**
     * write the Done record to the Howl Log
     */

    public void howlDoneLog (byte [] [] xaDnRec, XACommittingTx xaCmTx) throws LogException, Exception {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Done howl log");
        }
        xaLog.putDone (xaDnRec, xaCmTx);
    }
    
    /**
     * close Transaction Recovery Log
     */

    public void forget () throws LogException, Exception {
        
        if (!Current.getCurrent().getDefaultRecovery()) {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("JOTM Recovery is disabled");
            }  
        } else {
            howlCloseLog();
        }

   }

}
